import click
import qrcode
import rich

from decentnet.interface.alias_resolver import AliasResolver
from decentnet.modules.cryptography.asymetric import AsymCrypt
from decentnet.modules.key_util.key_manager import KeyManager


@click.group()
def key():
    pass


@key.command()
@click.option('--private-key-file', '-p', default=None,
              help='Filename for the private key')
@click.option('--public-key-file', '-u', default=None,
              help='Filename for the public key')
@click.option("--description", "-d", default="", help="Description of the key")
@click.option("--sign", "-s", default=False, type=bool, help="Signing keys or Encryption",
              is_flag=True)
@click.option("--alias", "-a", default=None, help="Alias of the key", type=str)
def generate(private_key_file, public_key_file, description, sign, alias):
    """
    Generate SSH key pair.
    """
    generate_impl(description, private_key_file, public_key_file, sign, alias)


def generate_impl(description: str, private_key_file: str = None, public_key_file: str = None,
                  sign: bool = False, alias: str = None):
    if sign:
        private_key, o_public_key = KeyManager.generate_singing_key_pair()
        public_key = AsymCrypt.verifying_key_to_string(o_public_key)
    else:
        private_key, o_public_key = KeyManager.generate_encryption_key_pair()
        public_key = KeyManager.key_to_base64(o_public_key)

    private_key = KeyManager.key_to_base64(private_key)

    if private_key_file:
        with open(private_key_file, 'w') as f:
            f.write(private_key)
    if public_key_file:
        with open(public_key_file, 'w') as f:
            f.write(public_key)
    KeyManager.save_to_db(private_key, public_key, description, not sign, alias)
    rich.print("[green]Generated new key and saved to database[/green]")


@key.command()
@click.option("--alias-sign", "-s", required=True, help="Alias of the key for signing")
@click.option("--alias-enc", "-e", required=True, help="Alias of the key for encryption")
@click.option('--qr', is_flag=True, help='Generate a QR code for the keys')
def share(a_sign: str, a_enc: str, qr: bool):
    akey, _ = AliasResolver.get_key_by_alias(a_sign)
    aenc, _ = AliasResolver.get_key_by_alias(a_enc)
    if qr:
        keys_qr = qrcode.make(akey + aenc)
        keys_qr.show()
    else:
        print(f"{akey}.{aenc}")


@key.command("import")
@click.argument('private_key_path', type=click.Path(exists=True))
@click.argument('public_key_path', type=click.Path(exists=True))
def do_import(private_key_path, public_key_path):
    """
    Import SSH key pair from files.
    """
    private_key_obj, public_key_obj = KeyManager.import_ssh_key_pair(private_key_path,
                                                                     public_key_path)
    private_key, public_key = KeyManager.export_ssh_key_pair(private_key_obj,
                                                             public_key_obj)
    click.echo(f'Private Key:\n{private_key}\n')
    click.echo(f'Public Key:\n{public_key}\n')



import logging
from typing import Optional

import cbor2
from ecdsa import VerifyingKey

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.modules.cryptography.asymetric import AsymCrypt
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


class Serializer:
    @staticmethod
    def serialize_data(
            public_key: VerifyingKey | str,
            signature: bytes,
            encrypted_data: bytes,
            target_pub_key: Optional[str] = None,
            cmd: Optional[int] = None,
            command_signature: Optional[bytes] = None,
            command_public_key: Optional[VerifyingKey] = None
    ) -> bytes:
        """Serializes public key, signature, and encrypted data and returns them as separate bytes.

        :returns: utf-8 encoded JSON
        """
        # Convert public_key to string if it's not already
        public_key_str = (
            public_key if isinstance(public_key, str)
            else AsymCrypt.verifying_key_to_string(public_key)
        )

        out = {
            "pub": public_key_str,
            "sig": signature,
            "data": encrypted_data,
            "target": target_pub_key
        }

        # Include command-related data only if cmd and command_signature are provided
        if cmd and command_signature:
            out.update({
                "cmd": cmd,
                "csig": command_signature,
                "cpub": AsymCrypt.verifying_key_to_string(command_public_key)
            })

        return cbor2.dumps(out)

    @staticmethod
    def deserialize_data(serialized_data: bytes) -> dict:
        """Deserializes the provided data and returns the public key, signature, and encrypted data."""
        deserialized_data = cbor2.loads(serialized_data)

        if deserialized_data.get("sig", None) is None:
            raise Exception(f"Corrupted data {deserialized_data}")

        return deserialized_data



import multiprocessing
import unittest

from decentnet.modules.blockchain.blockchain import Blockchain
from decentnet.modules.tcp.client import TCPClient
from decentnet.modules.tcp.server import TCPServer
from decentnet.modules.transfer.packager import Packager


class TestTCPCommunication(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.server = TCPServer('127.0.0.1', 8886)
        cls.server_process = multiprocessing.Process(target=cls.server.run)
        cls.server_process.start()

    @classmethod
    def tearDownClass(cls):
        cls.server.close()
        cls.server_process.terminate()

    def test_server_client_connect(self):
        client = TCPClient('127.0.0.1', 8886)
        bc = Blockchain("CONNECTED")

        data = Packager.pack(1,
                             bc.get_last(), None)
        response = client.send_message(data)

        self.assertEqual(response, "CONNECTED")

        client.close()


if __name__ == "__main__":
    unittest.main()



import logging
import zlib

import brotli
import lz4.frame

from decentnet.consensus.compress_params import COMPRESSION_LEVEL_LZ4, COMPRESSION_BLOCK_SIZE_LZ4
from decentnet.consensus.dev_constants import RUN_IN_DEBUG, COMPRESSION_LOG_LEVEL
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)
logger.setLevel(COMPRESSION_LOG_LEVEL)

setup_logger(RUN_IN_DEBUG, logger)


class CompressionWrapper:
    chunk_size = 1024

    @staticmethod
    def compress_brotli(data: bytes, quality: int = 11) -> bytes:
        """
        Compresses data using Brotli compression algorithm with the specified quality.

        Args:
            data (bytes): The data to compress.
            quality (int): The compression quality. Higher values result in better compression but slower speed. Defaults to 11.

        Returns:
            bytes: The compressed data.
        """
        return brotli.compress(data, quality=quality)

    @staticmethod
    def compress_zlib(data: bytes, level: int = 6) -> bytes:
        """
        Compresses data using zlib compression algorithm with the specified compression level.

        Args:
            data (bytes): The data to compress.
            level (int): The compression level. Higher values result in better compression but slower speed. Defaults to 6.

        Returns:
            bytes: The compressed data.
        """
        return zlib.compress(data, level=level)

    @staticmethod
    def decompress_brotli(compressed_data: bytes) -> bytes:
        """
        Decompresses data using Brotli decompression algorithm.

        Args:
            compressed_data (bytes): The compressed data to decompress.

        Returns:
            bytes: The decompressed data.
        """
        return brotli.decompress(compressed_data)

    @staticmethod
    def decompress_zlib(compressed_data: bytes) -> bytes:
        """
        Decompresses data using zlib decompression algorithm.

        Args:
            compressed_data (bytes): The compressed data to decompress.

        Returns:
            bytes: The decompressed data.
        """
        return zlib.decompress(compressed_data)

    @staticmethod
    def compress_lz4(data: bytes,
                     compression_level: int = COMPRESSION_LEVEL_LZ4,
                     block_size: int = COMPRESSION_BLOCK_SIZE_LZ4) -> bytes:
        """
        Compresses data using LZ4 compression algorithm with the specified compression level.

        Args:
            data (bytes): The data to compress.
            compression_level (int): The compression level (0-16), higher values result in better compression but slower speed. Defaults to 8.

        Returns:
            bytes: The compressed data.
        """
        data_size_before = len(data)
        logger.debug(f"Compressing data with LZ4 initial size {data_size_before} B")
        compressed = lz4.frame.compress(data, compression_level=compression_level,
                                        block_size=block_size)
        data_size_after = len(compressed)
        logger.debug(f"Compressed data to {data_size_after} B")
        if data_size_before < data_size_after:
            logger.debug(f"Disabled compression due to data incompatibility")
            return data
        return compressed

    @staticmethod
    def decompress_lz4(compressed_data: bytes) -> bytes:
        """
        Decompresses data using LZ4 decompression algorithm.

        Args:
            compressed_data (bytes): The compressed data to decompress.

        Returns:
            bytes: The decompressed data.
        """
        logger.debug("Decompressing data with LZ4")
        try:
            return lz4.frame.decompress(compressed_data)
        except RuntimeError:
            return compressed_data



from decentnet.modules.pow.pow_utils import PowUtils


def compute_pow(n_bits, hash_t, nonce):
    _bits = hash_t.diff.hash_len_chars * 8 - n_bits
    # Clone the original hash_t to avoid modifying the original object

    while PowUtils.get_bit_length(hash_t, nonce, hash_t.diff) > _bits:
        nonce += 1

    # Return the correct nonce but do not modify the original hash_t
    return nonce
